from collections import OrderedDict
from logging import Logger
from typing import Any, Dict, Generator, List

import requests
from requests.utils import requote_uri

from the_census._api.interface import (
    ICensusApiFetchService,
    ICensusApiSerializationService,
)
from the_census._api.models import GeographyItem
from the_census._config import Config
from the_census._exceptions import CensusDoesNotExistException, InvalidQueryException
from the_census._geographies.models import GeoDomain
from the_census._utils.chunk import chunk
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._variables.models import Group, GroupVariable, VariableCode

# we can query only 50 variables at a time, max
MAX_QUERY_SIZE = 50
API_URL_FORMAT = "https://api.census.gov/data/{0}/{1}/{2}"


class CensusApiFetchService(ICensusApiFetchService):
    _url: str
    _parser: ICensusApiSerializationService
    _config: Config
    _logger: Logger

    def __init__(
        self,
        config: Config,
        parser: ICensusApiSerializationService,
        logging_factory: ILoggerFactory,
    ) -> None:
        self._url = API_URL_FORMAT.format(config.year, config.dataset, config.survey)
        self._parser = parser
        self._config = config
        self._logger = logging_factory.getLogger(__name__)

    def healthcheck(self) -> None:
        res = requests.get(self._url + ".json")  # type: ignore

        if res.status_code in [404, 400]:
            msg = f"Data does not exist for dataset={self._config.dataset}; survey={self._config.survey}; year={self._config.year}"

            self._logger.exception(msg)

            raise CensusDoesNotExistException(msg)

        self._logger.debug("healthcheck OK")

    @timer
    def geography_codes(
        self, for_domain: GeoDomain, in_domains: List[GeoDomain] = []
    ) -> Any:

        for_clause = f"for={for_domain}"
        in_clauses = "&in=".join([str(parent) for parent in in_domains])

        querystring = f"?get=NAME&{for_clause}"
        if len(in_domains):
            querystring += f"&in={in_clauses}"  # type: ignore

        uri_querystring: str = requote_uri(querystring)

        return self._fetch(route=uri_querystring)

    @timer
    def group_data(self) -> Dict[str, Group]:
        groups_res: Dict[str, List[Dict[str, str]]] = self._fetch(route="/groups.json")

        return self._parser.parse_groups(groups_res)

    @timer
    def supported_geographies(self) -> OrderedDict[str, GeographyItem]:
        geogRes = self._fetch(route="/geography.json")

        return self._parser.parse_supported_geographies(geogRes)

    @timer
    def variables_for_group(self, group: str) -> List[GroupVariable]:
        res = self._fetch(route=f"/groups/{group}.json")

        return self._parser.parse_group_variables(res)

    @timer
    def all_variables(self) -> List[GroupVariable]:
        res = self._fetch("/variables.json")

        return self._parser.parse_group_variables(res)

    @timer
    def stats(
        self,
        variables_codes: List[VariableCode],
        for_domain: GeoDomain,
        in_domains: List[GeoDomain] = [],
    ) -> Generator[List[List[str]], None, None]:

        # we need the minus 1 since we're also querying name
        for codes in chunk(variables_codes, MAX_QUERY_SIZE - 1):
            code_str = ",".join(codes)
            var_str = "get=NAME" + f",{code_str}" if len(code_str) > 0 else ""

            domainStr = "for=" + str(for_domain)
            in_domainstr = "&".join([f"in={domain}" for domain in in_domains])

            if len(in_domainstr) > 0:
                domainStr += "&"  # type: ignore
                domainStr += in_domainstr  # type: ignore

            route = f"?{var_str}&{domainStr}"

            uri_route: str = requote_uri(route)

            # not doing any serializing here, because this is a bit more
            # complicated (we need to convert the stats to the appropriate
            # data types, [e.g., int, float] further up when we're working
            # with dataFrames; there's no real good way to do it down here)

            yield self._fetch(uri_route)

    def _fetch(self, route: str = "") -> Any:
        ampersand_or_question_mark = "&" if "?" in route else "?"
        url = (
            self._url
            + route
            + ampersand_or_question_mark
            + "key="
            + self._config.api_key
        )
        res = requests.get(url)  # type: ignore

        if res.status_code in [400, 404]:
            msg = f"Could not make query for route `{route}`"
            self._logger.exception(msg)
            raise InvalidQueryException(msg)

        if res.status_code == 204:  # no content
            msg = f"Received no content for query for route {route}"
            self._logger.info(msg)

            return []

        return res.json()  # type: ignore



from dataclasses import dataclass
from functools import cache
from typing import Any, Dict, List, cast

import pandas
import pandas as pd
import requests
from tqdm.notebook import tqdm

URL = "https://api.census.gov/data.json"

# documentation: https://www2.census.gov/programs-surveys/acs/tech_docs/subject_definitions/


@dataclass(frozen=True)
class _DatasetsRes:
    year: int
    dataset: List[str]
    is_aggregate: bool
    title: str
    description: str

    @classmethod
    def from_json(cls, jsonRes: Dict[str, Any]):
        return cls(
            cast(int, jsonRes.get("c_vintage")),
            cast(List[str], jsonRes.get("c_dataset")),
            cast(bool, jsonRes.get("c_isAggregate")),
            cast(str, jsonRes.get("title")),
            cast(str, jsonRes.get("description")),
        )


def list_available_datasets() -> pd.DataFrame:
    return __list_available_datasets()


@cache
def __list_available_datasets() -> pd.DataFrame:
    res: Dict[str, Any] = requests.get(URL).json()  # type: ignore
    dataset_dicts: List[Dict[str, str]] = []

    available_datasets: List[_DatasetsRes] = [
        _DatasetsRes.from_json(datasetJson) for datasetJson in res["dataset"]
    ]

    for dataset in cast(List[_DatasetsRes], tqdm(available_datasets)):
        # these won't play nice with the tool
        if not dataset.is_aggregate:
            continue

        dataset_type = ""
        survey_type = ""
        if len(dataset.dataset) > 0:
            dataset_type = dataset.dataset[0]
            if len(dataset.dataset) > 1:
                survey_type = "/".join(dataset.dataset[1:])

        dataset_dicts.append(
            cast(
                Dict[str, str],
                dict(
                    year=dataset.year,
                    name=dataset.title,
                    description=dataset.description,
                    dataset=dataset_type,
                    survey=survey_type,
                ),
            )
        )

    pandas.set_option("display.max_colwidth", None)  # type: ignore

    return (
        pd.DataFrame(dataset_dicts)
        .sort_values(by=["year", "name"], ascending=[False, True])
        .reset_index(drop=True)
    )



from dataclasses import dataclass

CACHE_DIR = "cache"


@dataclass(frozen=True)
class Config:
    """
    Stores basic information so that we don't need to
    pass the same variables around all of the time
    """

    year: int = 2020
    dataset: str = "acs"
    survey: str = "acs1"
    cache_dir: str = CACHE_DIR
    should_load_from_existing_cache: bool = False
    should_cache_on_disk: bool = False
    replace_column_headers: bool = False
    api_key: str = ""



from abc import ABC, abstractmethod
from typing import Generic, ItemsView, KeysView, TypeVar, ValuesView

from the_census._variables.models import Group, GroupCode, GroupVariable

ValueType = TypeVar("ValueType")
ItemType = TypeVar("ItemType")


class ICodeSet(ABC, Generic[ItemType, ValueType]):
    def __init__(self, *items: ItemType) -> None:
        self.add(*items)

    @abstractmethod
    def add(self, *items: ItemType):
        ...

    def __len__(self) -> int:
        return len(self.__dict__)

    def names(self) -> KeysView[str]:
        return self.__dict__.keys()

    def values(self) -> ValuesView[ValueType]:
        return self.__dict__.values()

    def items(self) -> ItemsView[str, ValueType]:
        return self.__dict__.items()

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __str__(self) -> str:
        return self.__repr__()

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, type(self)):
            return False

        if len(o) != len(self):
            return False

        for k, v in self.items():
            if not hasattr(o, k):
                return False
            if v != getattr(o, k):
                return False

        return True

    def __ne__(self, o: object) -> bool:
        return not self == o


class VariableSet(ICodeSet[GroupVariable, GroupVariable]):
    def add(self, *items: GroupVariable):
        filteredItems = [item for item in items if item not in self.values()]

        entries = {
            f"{item.cleaned_name}_{item.group_code}": item for item in filteredItems
        }

        self.__dict__.update(entries)


class GroupSet(ICodeSet[Group, GroupCode]):
    def add(self, *items: Group):
        itemNames = [item.cleaned_name for item in items]

        for item in items:
            if item.code in self.values():
                continue

            cleaned_name_freq = len(
                [1 for name in itemNames if item.cleaned_name == name]
            )

            if cleaned_name_freq > 1 or item.cleaned_name in self.__dict__:
                self.__dict__.update({f"{item.cleaned_name}_{item.code}": item.code})
            else:
                self.__dict__.update({item.cleaned_name: item.code})



from typing import Generator, List, TypeVar

_T = TypeVar("_T")


def chunk(items: List[_T], n: int) -> Generator[List[_T], None, None]:
    """
    Splits a list into `n`-sized chunks

    Args:
        items (List[_T]): the list
        n (int): the chunk size

    Yields:
        Generator[List[_T], None, None]: Generator of each chunk
    """
    for i in range(0, len(items), n):
        yield items[i : i + n]



# from dataclasses import dataclass
from dataclasses import dataclass, field
from typing import Any, Dict, Literal, NewType

from the_census._utils.clean_variable_name import clean_variable_name

VariableCode = NewType("VariableCode", str)
GroupCode = NewType("group_code", str)


@dataclass(frozen=True)
class Group:
    """
    DTO for the group retrieved from the API
    """

    code: GroupCode
    description: str
    variables: str = field(default="")
    cleaned_name: str = field(default="")

    @classmethod
    def from_json(cls, jsonDict: Dict[str, str]):
        code = jsonDict.get("name", "")
        description = jsonDict.get("description", "")
        variables = jsonDict.get("variables", "")

        return cls(
            GroupCode(code), description, variables, clean_variable_name(description)
        )

    @classmethod
    def from_df_record(cls, record: Dict[str, Any]):
        return cls(
            GroupCode(record["code"]),
            record["description"],
            cleaned_name=record["cleaned_name"],
        )


@dataclass
class GroupVariable:
    """
    DTO for a variable retrieved from the API
    """

    code: VariableCode
    group_code: GroupCode
    group_concept: str
    name: str
    limit: int = field(default=0)
    predicate_only: bool = field(default=False)
    predicate_type: Literal["string", "int", "float"] = field(default="string")
    cleaned_name: str = field(default="")

    @classmethod
    def from_json(cls, code: str, jsonData: Dict[Any, Any]):
        group_code = jsonData.get("group", "")
        group_concept = jsonData.get("concept", "")
        label = jsonData.get("label", "")
        limit = jsonData.get("limit", 0)
        predicate_only = jsonData.get("predicateOnly", False)
        predicate_type = jsonData.get("predicateType", "string")
        cleaned_name = clean_variable_name(label)

        return cls(
            VariableCode(code),
            GroupCode(group_code),
            group_concept,
            label,
            limit,
            predicate_only,
            predicate_type,
            cleaned_name,
        )

    @classmethod
    def from_df_record(cls, record: Dict[str, Any]):
        return cls(
            VariableCode(record["code"]),
            GroupCode(record["group_code"]),
            record["group_concept"],
            record["name"],
            record["limit"],
            record["predicate_only"],
            record["predicate_type"],
            record["cleaned_name"],
        )

    def __hash__(self) -> int:
        return hash(self.code)



from typing import List

import pandas as pd

from the_census._api.fetch import ICensusApiFetchService
from the_census._geographies.interface import IGeographyRepository
from the_census._geographies.models import GeoDomain, GeoDomainTypes, SupportedGeoSet
from the_census._stats.interface import ICensusStatisticsService
from the_census._variables.models import GroupCode, VariableCode
from the_census._variables.repository.interface import IVariableRepository
from the_census._variables.repository.models import GroupSet, VariableSet
from the_census._variables.search.interface import IVariableSearchService


class CensusClient:
    """
    The Census client. Using a layer of indirection to hide
    additional properties from consumers.
    """

    _variable_repo: IVariableRepository[pd.DataFrame]
    _variableSearch: IVariableSearchService[pd.DataFrame]
    _stats: ICensusStatisticsService[pd.DataFrame]
    _geo_repo: IGeographyRepository[pd.DataFrame]

    def __init__(
        self,
        variableRepo: IVariableRepository[pd.DataFrame],
        variableSearch: IVariableSearchService[pd.DataFrame],
        stats: ICensusStatisticsService[pd.DataFrame],
        api: ICensusApiFetchService,
        geoRepo: IGeographyRepository[pd.DataFrame],
    ) -> None:
        self._variable_repo = variableRepo
        self._variableSearch = variableSearch
        self._stats = stats
        self._geo_repo = geoRepo

        # if this healthcheck fails, it will throw, and we
        # won't instantiate the client
        api.healthcheck()

    # search
    def search_groups(self, regex: str) -> pd.DataFrame:
        return self._variableSearch.search_groups(regex).copy(deep=True)

    def search_variables(
        self,
        regex: str,
        *in_groups: GroupCode,
    ) -> pd.DataFrame:
        return self._variableSearch.search_variables(regex, *in_groups).copy(deep=True)

    # repo
    def get_geography_codes(
        self, for_domain: GeoDomainTypes, *in_domains: GeoDomainTypes
    ) -> pd.DataFrame:
        return self._geo_repo.get_geography_codes(
            GeoDomain._from(for_domain),
            *[GeoDomain._from(in_domain) for in_domain in in_domains],
        ).copy(deep=True)

    def get_groups(self) -> pd.DataFrame:
        return self._variable_repo.get_groups().copy(deep=True)

    def get_variables_by_group(self, *groups: GroupCode) -> pd.DataFrame:
        return self._variable_repo.get_variables_by_group(*groups).copy(deep=True)

    def get_all_variables(self) -> pd.DataFrame:
        return self._variable_repo.get_all_variables().copy(deep=True)

    def get_supported_geographies(self) -> pd.DataFrame:
        return self._geo_repo.get_supported_geographies().copy(deep=True)

    def get_stats(
        self,
        variables_to_query: List[VariableCode],
        for_domain: GeoDomainTypes,
        *in_domains: GeoDomainTypes,
    ) -> pd.DataFrame:
        return self._stats.get_stats(
            variables_to_query,
            GeoDomain._from(for_domain),
            *[GeoDomain._from(in_domain) for in_domain in in_domains],
        ).copy(deep=True)

    # helpers

    # property variables for Jupyter notebook usage

    @property
    def variables(self) -> VariableSet:
        return self._variable_repo.variables

    @property
    def groups(self) -> GroupSet:
        return self._variable_repo.groups

    @property
    def supported_geographies(self) -> SupportedGeoSet:
        return self._geo_repo.supported_geographies



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from the_census._variables.models import GroupCode

_T = TypeVar("_T")


class IVariableSearchService(ABC, Generic[_T]):
    """
    Handles searching through stored variables
    """

    @abstractmethod
    def search_groups(self, regex: str) -> _T:
        ...

    @abstractmethod
    def search_variables(
        self,
        regex: str,
        *in_groups: GroupCode,
    ) -> _T:
        ...



from functools import cache
from logging import Logger
from typing import List, Tuple, cast

import pandas as pd

from the_census._api.interface import ICensusApiFetchService
from the_census._data_transformation.interface import ICensusDataTransformer
from the_census._geographies.interface import IGeographyRepository
from the_census._geographies.models import GeoDomain, SupportedGeoSet
from the_census._persistence.interface import ICache
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._utils.unique import get_unique

SUPPORTED_GEOS_FILE = "supported_geographies.csv"


class GeographyRepository(IGeographyRepository[pd.DataFrame]):
    _cache: ICache[pd.DataFrame]
    _api: ICensusApiFetchService
    _transformer: ICensusDataTransformer[pd.DataFrame]
    _logger: Logger

    def __init__(
        self,
        cache: ICache[pd.DataFrame],
        api: ICensusApiFetchService,
        transformer: ICensusDataTransformer[pd.DataFrame],
        logger_factory: ILoggerFactory,
    ) -> None:
        self._cache = cache
        self._api = api
        self._transformer = transformer
        self._logger = logger_factory.getLogger(__name__)

        self._supported_geographies = SupportedGeoSet()

        self.__populate_repository()

    @timer
    def get_geography_codes(
        self, for_domain: GeoDomain, *in_domains: GeoDomain
    ) -> pd.DataFrame:
        return self.__get_geography_codes(
            for_domain, in_domains=tuple(get_unique(in_domains))
        )

    @cache
    def __get_geography_codes(
        self, for_domain: GeoDomain, in_domains: Tuple[GeoDomain, ...] = ()
    ) -> pd.DataFrame:
        self._logger.debug(f"getting geography codes for {for_domain} in {in_domains}")
        res = self._api.geography_codes(for_domain, list(in_domains))

        if len(res) == 0:
            return pd.DataFrame()

        df = self._transformer.geography_codes(res)

        return df

    @timer
    def get_supported_geographies(self) -> pd.DataFrame:
        return self.__get_supported_geographies()

    @cache
    def __get_supported_geographies(self) -> pd.DataFrame:
        self._logger.debug("getting supported geographies")

        df = self._cache.get(SUPPORTED_GEOS_FILE)

        if df is None:
            df = pd.DataFrame()

        if df.empty:
            res = self._api.supported_geographies()

            if len(res) == 0:
                return pd.DataFrame()

            df = self._transformer.supported_geographies(res)

            self._cache.put(SUPPORTED_GEOS_FILE, df)

        self._supported_geographies.add(*df["name"].tolist())

        return df

    def __populate_repository(self) -> None:
        self._logger.debug("Trying to populate geography repository")

        df = self._cache.get(SUPPORTED_GEOS_FILE)

        if df is None or df.empty:
            self._logger.debug("Could not populate geography repository")
            return

        self._supported_geographies.add(*set(cast(List[str], df["name"].tolist())))

        self._logger.debug("Populated geography repository")



from dataclasses import dataclass, field
from typing import Tuple, Union

from the_census._utils.clean_variable_name import clean_variable_name

GeoDomainTypes = Union["GeoDomain", Tuple[str, str], Tuple[str]]


@dataclass(frozen=True)
class GeoDomain:
    name: str
    code_or_wildcard: str = field(default="*")

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"{self.name}:{self.code_or_wildcard}"

    @classmethod
    def _from(
        cls,
        geo_domain: Union[
            "GeoDomain",
            Tuple[
                str,
                str,
            ],
            Tuple[str],
        ],
    ) -> "GeoDomain":
        if isinstance(geo_domain, GeoDomain):
            return cls(geo_domain.name, geo_domain.code_or_wildcard)
        else:
            return cls(*geo_domain)


class SupportedGeoSet:
    def __init__(self) -> None:
        super().__init__()

    def add(self, *geo_name: str):
        mapping = {clean_variable_name(name): name for name in geo_name}
        self.__dict__.update(mapping)



from functools import cache
from logging import Logger
from typing import Any, Dict, List, Set, Tuple

import pandas as pd

from the_census._api.interface import ICensusApiFetchService
from the_census._data_transformation.interface import ICensusDataTransformer
from the_census._exceptions import EmptyRepositoryException
from the_census._geographies.interface import IGeographyRepository
from the_census._geographies.models import GeoDomain
from the_census._stats.interface import ICensusStatisticsService
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._utils.unique import get_unique
from the_census._variables.models import VariableCode
from the_census._variables.repository.interface import IVariableRepository


class CensusStatisticsService(ICensusStatisticsService[pd.DataFrame]):
    _api: ICensusApiFetchService
    _transformer: ICensusDataTransformer[pd.DataFrame]
    _variable_repo: IVariableRepository[pd.DataFrame]
    _geo_repo: IGeographyRepository[pd.DataFrame]
    _logger: Logger

    def __init__(
        self,
        api: ICensusApiFetchService,
        transformer: ICensusDataTransformer[pd.DataFrame],
        variableRepo: IVariableRepository[pd.DataFrame],
        geoRepo: IGeographyRepository[pd.DataFrame],
        loggerFactory: ILoggerFactory,
    ) -> None:
        self._api = api
        self._transformer = transformer
        self._variable_repo = variableRepo
        self._geo_repo = geoRepo
        self._logger = loggerFactory.getLogger(__name__)

    @timer
    def get_stats(
        self,
        variables_to_query: List[VariableCode],
        for_domain: GeoDomain,
        *in_domains: GeoDomain,
    ) -> pd.DataFrame:

        return self.__get_stats(
            variables_to_query=tuple(get_unique(variables_to_query)),
            for_domain=for_domain,
            in_domains=tuple(get_unique(in_domains)),
        )

    @cache
    def __get_stats(
        self,
        variables_to_query: Tuple[VariableCode],
        for_domain: GeoDomain,
        in_domains: Tuple[GeoDomain],
    ) -> pd.DataFrame:

        pullStats = lambda: self._api.stats(
            list(variables_to_query), for_domain, list(in_domains)
        )

        apiResults: List[List[List[str]]] = [res for res in pullStats()]

        (
            column_headers,
            type_conversions,
        ) = self._get_variable_names_and_type_conversions(set(variables_to_query))

        geo_domains_queried = [for_domain] + list(in_domains)

        supported_geos = self._geo_repo.get_supported_geographies()

        df = self._transformer.stats(
            apiResults,
            type_conversions,
            geo_domains_queried,
            column_headers,
            supported_geos,
        )

        return df

    def _get_variable_names_and_type_conversions(
        self, variables_to_query: Set[VariableCode]
    ) -> Tuple[Dict[VariableCode, str], Dict[str, Any]]:

        relevant_variables = {
            variable.code: variable
            for variable in self._variable_repo.variables.values()
            if variable.code in variables_to_query
        }
        if len(relevant_variables) != len(variables_to_query):
            msg = f"Queried {len(variables_to_query)} variables, but found only {len(relevant_variables)} in repository"

            self._logger.exception(msg)

            raise EmptyRepositoryException(msg)

        hasDuplicateNames = len(
            {v.cleaned_name for v in relevant_variables.values()}
        ) < len(variables_to_query)

        type_conversions: Dict[str, Any] = {}
        column_headers: Dict[VariableCode, str] = {}
        for k, v in relevant_variables.items():
            if v.predicate_type in ["int", "float"]:
                type_conversions.update({k: float})

            cleanedVarName = v.cleaned_name
            if hasDuplicateNames:
                cleanedVarName += f"_{v.group_code}"

            column_headers.update({k: cleanedVarName})

        return column_headers, type_conversions



import os
from functools import cache
from logging import Logger
from pathlib import Path
from typing import Tuple, cast

import pandas as pd
from tqdm.notebook import tqdm

from the_census._api.interface import ICensusApiFetchService
from the_census._data_transformation.interface import ICensusDataTransformer
from the_census._persistence.interface import ICache
from the_census._utils.log.factory import ILoggerFactory
from the_census._utils.timer import timer
from the_census._utils.unique import get_unique
from the_census._variables.models import Group, GroupCode, GroupVariable
from the_census._variables.repository.interface import IVariableRepository
from the_census._variables.repository.models import GroupSet, VariableSet

GROUPS_FILE = "groups.csv"
SUPPORTED_GEOS_FILE = "supported_geographies.csv"

VARIABLES_DIR = "variables"


class VariableRepository(IVariableRepository[pd.DataFrame]):

    _cache: ICache[pd.DataFrame]
    _api: ICensusApiFetchService
    _transformer: ICensusDataTransformer[pd.DataFrame]
    _logger: Logger

    def __init__(
        self,
        cache: ICache[pd.DataFrame],
        transformer: ICensusDataTransformer[pd.DataFrame],
        api: ICensusApiFetchService,
        logger_factory: ILoggerFactory,
    ):
        self._cache = cache
        self._api = api
        self._transformer = transformer
        self._logger = logger_factory.getLogger(__name__)

        # these are inherited from the base class
        self._variables = VariableSet()
        self._groups = GroupSet()

        self.__populate_repository()

    @timer
    def get_groups(self) -> pd.DataFrame:
        return self.__get_groups()

    @cache
    def __get_groups(self) -> pd.DataFrame:
        df = self._cache.get(GROUPS_FILE)
        if df is None:
            df = pd.DataFrame()

        if df.empty:
            res = self._api.group_data()

            if len(res) == 0:
                return pd.DataFrame()

            df = self._transformer.groups(res)

            self._cache.put(GROUPS_FILE, df)

        groups = [Group.from_df_record(rec) for rec in df.to_dict("records")]
        self._groups.add(*groups)

        return df.drop(columns=["cleaned_name"])  # type: ignore

    @timer
    def get_variables_by_group(self, *groups: GroupCode) -> pd.DataFrame:
        return self.__get_variables_by_group(tuple(get_unique(groups)))

    @cache
    def __get_variables_by_group(self, groups: Tuple[GroupCode, ...]) -> pd.DataFrame:
        all_vars = pd.DataFrame()

        for group in tqdm(groups):  # type: ignore
            resource = f"{VARIABLES_DIR}/{group}.csv"

            df = self._cache.get(resource)
            if df is None:
                df = pd.DataFrame()

            if not df.empty:
                if all_vars.empty:
                    all_vars = df
                else:
                    all_vars = all_vars.append(df, ignore_index=True)  # type: ignore
            else:
                res = self._api.variables_for_group(cast(GroupCode, group))

                if len(res) == 0:
                    continue

                df = self._transformer.variables(res)

                self._cache.put(resource, df)

                if all_vars.empty:
                    all_vars = df
                else:
                    all_vars = all_vars.append(df, ignore_index=True)

        if all_vars.empty:
            return pd.DataFrame()

        for record in all_vars.to_dict("records"):
            var = GroupVariable.from_df_record(record)
            self._variables.add(var)

        return all_vars.drop(columns=["cleaned_name"])  # type: ignore

    @timer
    def get_all_variables(self) -> pd.DataFrame:
        return self.__get_all_variables()

    @cache
    def __get_all_variables(self) -> pd.DataFrame:
        self._logger.info("This is a costly operation, and may take time")

        all_variables = self._api.all_variables()

        if len(all_variables) == 0:
            return pd.DataFrame()

        df = self._transformer.variables(all_variables)

        for GroupCode, variables in df.groupby(["group_code"]):  # type: ignore
            varDf = cast(pd.DataFrame, variables)

            if not self._cache.put(f"{VARIABLES_DIR}/{GroupCode}.csv", varDf):
                # we don't need to update `self._variables` in this case
                continue

            variables = [
                GroupVariable.from_df_record(record)
                for record in varDf.to_dict("records")
            ]
            self._variables.add(*variables)

        return df.drop(columns=["cleaned_name"])  # type: ignore

    def __populate_repository(self) -> None:
        self._logger.debug("populating repository")

        groups_df = self._cache.get(GROUPS_FILE)

        if groups_df is not None:
            groups = [
                Group.from_df_record(record) for record in groups_df.to_dict("records")
            ]
            self._logger.debug(f"adding groups {[group.code for group in groups]}")
            self._groups.add(*groups)

        variables_path = f"{self._cache.cache_path}/{VARIABLES_DIR}"

        if not Path(variables_path).exists():
            self._logger.debug("no variables")
            return

        for file in os.listdir(variables_path):
            cache_res = self._cache.get(f"{VARIABLES_DIR}/{file}")
            variable_df = cache_res if cache_res is not None else pd.DataFrame()

            self._logger.debug(f"adding variables from {file}")

            variables = [
                GroupVariable.from_df_record(record)
                for record in variable_df.to_dict("records")
            ]
            self._variables.add(*variables)



from abc import abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Generic, List, TypeVar

from the_census._api.models import GeographyItem
from the_census._geographies.models import GeoDomain
from the_census._variables.models import Group, GroupVariable, VariableCode

T = TypeVar("T")


class ICensusDataTransformer(Generic[T]):
    """
    This takes care of converting parsed API data
    into more meaningfully consumable data (e.g., DataFrames)
    """

    @abstractmethod
    def supported_geographies(
        self, supported_geos: OrderedDict[str, GeographyItem]
    ) -> T:
        ...

    @abstractmethod
    def geography_codes(self, geo_codes: List[List[str]]) -> T:
        ...

    @abstractmethod
    def groups(self, groups: Dict[str, Group]) -> T:
        ...

    @abstractmethod
    def variables(self, variables: List[GroupVariable]) -> T:
        ...

    @abstractmethod
    def stats(
        self,
        results: List[List[List[str]]],
        type_conversions: Dict[str, Any],
        geo_domains_queried: List[GeoDomain],
        column_headers: Dict[VariableCode, str],
        supported_geos: T,
    ) -> T:
        """Parses stats data.

        This one get can get a bit complicated. Here's why:

        For stats, we might need to make batches API calls. This means
        that, when storing all of this in a DataFrame, we'll need to join
        on the geography IDs. This is why we pass in the geography domains
        which the stats service used to get the stats to the transformer.

        We can't rely on `results` for parsing out the geography columns,
        since `results` can have anywhere between 1 and 50 variables; and
        there's no way to know what portion of `results` is variables,
        and what portion is geography information.

        Args:
            results (List[List[List[str]]]): from the API
            type_conversions (Dict[str, Any]): for converting the data (which may all be strings) to
                floats if necessary, otherwise strings
            geo_domains_queried (List[str]): by the stats service
            column_headers (Dict[VariableCode, str]): the column headers with cleaned names
            supportedGoes (T): we need to pass this in (as opposed to DI-ing the Geography
                repo), since that would result in a circular dependency otherwise

        Returns:
            T: [description]
        """
        ...



import logging
import time
from typing import Any, Callable, Dict, TypeVar, cast

_Func = TypeVar("_Func", bound=Callable[..., Any])


def timer(func: _Func) -> _Func:
    def wrapper(*args: Any, **kwargs: Dict[Any, Any]) -> Any:
        start_time = time.perf_counter()

        retval = func(*args, **kwargs)

        end_time = time.perf_counter()

        elapsed_ms = (end_time - start_time) * 1000

        logging.getLogger(__name__).debug(
            f"[{func.__qualname__}] - duration: {elapsed_ms:.2f}ms"
        )

        return retval

    return cast(_Func, wrapper)  # type: ignore



def clean_variable_name(variable_name: str) -> str:
    """
    Removes punctuation/spaces from variable names

    Args:
        variable_name (str)

    Returns:
        str: without punctuation
    """

    full_name = ""

    for i, chunk in enumerate(variable_name.split("!!")):
        new_name = ""
        for sub_chunk in (
            chunk.replace(",", " ")
            .replace("-", " ")
            .replace(".", " ")
            .replace("(", " ")
            .replace(")", " ")
            .replace("[", " ")
            .replace("]", " ")
            .replace("/", " slash ")
            .split(" ")
        ):
            new_name += (
                sub_chunk.capitalize()
                .replace(" ", "")
                .replace(":", "")
                .replace("'", "")
            )

        if i == 0:
            full_name += new_name
        else:
            full_name += f"_{new_name}"

    return full_name



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from the_census._geographies.models import GeoDomain, SupportedGeoSet

T = TypeVar("T")


class IGeographyRepository(ABC, Generic[T]):
    """
    Gets and stores all geography information
    """

    _supported_geographies: SupportedGeoSet

    @abstractmethod
    def get_supported_geographies(self) -> T:
        ...

    @abstractmethod
    def get_geography_codes(self, for_domain: GeoDomain, *in_domains: GeoDomain) -> T:
        ...

    @property
    def supported_geographies(self) -> SupportedGeoSet:
        return self._supported_geographies



import logging
import sys

from the_census._utils.log.filters import ModuleFilter

DEFAULT_LOG_FILE = "census.log"


def configureLogger(log_file: str, datasetName: str) -> None:
    """
    sets up logger for the project

    Args:
        log_file (str): the name of the file that log output will be sent to
    """

    logFormat = "[{0} %(levelname)s] %(asctime)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s".format(
        datasetName
    )
    dateFormat = "%Y-%m-%d %H:%M:%S%z"

    logger = logging.getLogger("census")
    logger.setLevel(logging.NOTSET)

    streamHandler = logging.StreamHandler(sys.stdout)
    streamHandler.setLevel(logging.INFO)
    formatter = logging.Formatter(logFormat, datefmt=dateFormat)
    streamHandler.setFormatter(formatter)
    logger.addHandler(streamHandler)

    rootHandler = logging.FileHandler(log_file)
    rootHandler.setLevel(logging.NOTSET)
    rootHandler.addFilter(ModuleFilter())
    rootHandler.setFormatter(formatter)

    logging.basicConfig(level=logging.DEBUG, format=logFormat, handlers=[rootHandler])



from typing import Collection, List, Set, TypeVar

T = TypeVar("T")


def get_unique(items: Collection[T]) -> List[T]:
    """
    Get all unique elements in a list

    Args:
        items (List[T])

    Returns:
        List[T]: list with distinct items
    """

    unique_list: List[T] = []
    seen_items: Set[T] = set()

    for item in items:
        if item in seen_items:
            continue
        seen_items.add(item)
        unique_list.append(item)

    return unique_list



from collections import OrderedDict
from typing import Any, Dict, List

from the_census._api.interface import ICensusApiSerializationService
from the_census._api.models import (
    GeographyClauseSet,
    GeographyItem,
    GeographyResponseItem,
)
from the_census._utils.timer import timer
from the_census._variables.models import Group, GroupVariable


class ApiSerializationService(ICensusApiSerializationService):
    @timer
    def parse_group_variables(self, group_variables: Any) -> List[GroupVariable]:
        if len(group_variables) == 0:
            return []

        variables: List[GroupVariable] = []
        for varCode, varData in group_variables["variables"].items():
            groupVar = GroupVariable.from_json(varCode, varData)
            variables.append(groupVar)

        return variables

    @timer
    def parse_supported_geographies(
        self,
        supported_geos_response: Any,
    ) -> OrderedDict[str, GeographyItem]:

        supported_geographies: Dict[str, GeographyItem] = {}

        if len(supported_geos_response) == 0:
            return OrderedDict(supported_geographies)

        fips = [
            GeographyResponseItem.from_json(fip)
            for fip in supported_geos_response["fips"]
        ]

        for fip in fips:
            varName = fip.name
            requirements = fip.requires or []
            wildcards = fip.wildcard or []
            nonWildcardableRequirements = list(
                filter(lambda req: req not in wildcards, fip.requires)
            )

            withAllCodes = GeographyClauseSet.makeSet(
                for_clause=f"{varName}:CODE",
                in_clauses=[f"{requirement}:CODE" for requirement in requirements],
            )

            withWithCardForVar = GeographyClauseSet.makeSet(
                for_clause=f"{varName}:*",
                in_clauses=[
                    f"{requirement}:CODE" for requirement in nonWildcardableRequirements
                ],
            )

            withWildCardedRequirements = GeographyClauseSet.makeSet(
                for_clause=f"{varName}:*",
                in_clauses=[
                    f"{requirement}:CODE" for requirement in nonWildcardableRequirements
                ]
                + [f"{wildcard}:*" for wildcard in wildcards],
            )

            supported_geographies[varName] = GeographyItem.makeItem(
                name=varName,
                hierarchy=fip.geoLevelDisplay,
                clauses=[withAllCodes, withWithCardForVar, withWildCardedRequirements],
            )

        return OrderedDict(
            sorted(supported_geographies.items(), key=lambda t: t[1].hierarchy)
        )

    @timer
    def parse_groups(
        self, groups_res: Dict[str, List[Dict[str, str]]]
    ) -> Dict[str, Group]:

        if len(groups_res) == 0:
            return {}

        return {
            Group.from_json(group).code: Group.from_json(group)
            for group in groups_res["groups"]
        }



from dataclasses import dataclass, field
from typing import Dict, List, Tuple, cast

from the_census._utils.unique import get_unique


@dataclass(frozen=True)
class GeographyClauseSet:
    for_clause: str
    in_clauses: Tuple[str, ...]

    @classmethod
    def makeSet(cls, for_clause: str, in_clauses: List[str]):
        return cls(for_clause, tuple(get_unique(in_clauses)))

    def __repr__(self) -> str:
        return "\n".join([self.for_clause] + list(self.in_clauses))

    def __str__(self) -> str:
        return self.__repr__()


@dataclass(frozen=True)
class GeographyItem:
    name: str
    hierarchy: str
    clauses: Tuple[GeographyClauseSet, ...]

    @classmethod
    def makeItem(cls, name: str, hierarchy: str, clauses: List[GeographyClauseSet]):
        return cls(name, hierarchy, tuple(get_unique(clauses)))


@dataclass(frozen=True)
class GeographyResponseItem:
    name: str
    geoLevelDisplay: str
    referenceData: str
    requires: List[str]
    wildcard: List[str]
    optionalWithWCFor: str = field(default="")

    @classmethod
    def from_json(cls, jsonRes: Dict[str, str]):
        return cls(
            jsonRes.get("name", ""),
            jsonRes.get("geoLevelDisplay", ""),
            jsonRes.get("referenceDate", ""),
            cast(List[str], jsonRes.get("requires", [])),
            cast(List[str], jsonRes.get("wildcard", [])),
            jsonRes.get("optionalWithWCFor", ""),
        )



from collections import OrderedDict
from typing import Any, Dict, List, Tuple, Union, cast

import pandas as pd

from the_census._api.models import GeographyItem
from the_census._config import Config
from the_census._data_transformation.interface import ICensusDataTransformer
from the_census._geographies.models import GeoDomain
from the_census._utils.timer import timer
from the_census._variables.models import Group, GroupVariable, VariableCode


class CensusDataTransformer(ICensusDataTransformer[pd.DataFrame]):

    _config: Config

    def __init__(self, config: Config) -> None:

        self._config = config

    @timer
    def supported_geographies(
        self, supported_geos: OrderedDict[str, GeographyItem]
    ) -> pd.DataFrame:
        values_flattened: List[Dict[str, str]] = []
        for geo_item in supported_geos.values():
            for clause in geo_item.clauses:
                values_flattened.append(
                    {
                        "name": geo_item.name,
                        "hierarchy": geo_item.hierarchy,
                        "for": clause.for_clause,
                        "in": ",".join(clause.in_clauses),
                    }
                )

        return pd.DataFrame(values_flattened)

    @timer
    def geography_codes(self, geo_codes: List[List[str]]) -> pd.DataFrame:
        return (
            pd.DataFrame(geo_codes[1:], columns=geo_codes[0])
            .sort_values(by=geo_codes[0][1:])
            .reset_index(drop=True)
        )

    @timer
    def groups(self, groups: Dict[str, Group]) -> pd.DataFrame:
        group_list = []
        for code, groupObj in groups.items():
            group_dict = {
                "code": code,
                "description": groupObj.description,
                "cleaned_name": groupObj.cleaned_name,
            }
            group_list.append(group_dict)

        return pd.DataFrame(group_list).sort_values(by="code")

    @timer
    def variables(self, variables: List[GroupVariable]) -> pd.DataFrame:
        variable_dict_list: List[Dict[str, Union[str, int, bool]]] = []

        for variable in variables:
            variable_dict_list.append(
                {
                    "code": variable.code,
                    "group_code": variable.group_code,
                    "group_concept": variable.group_concept,
                    "name": variable.name,
                    "predicate_type": variable.predicate_type,
                    "predicate_only": variable.predicate_only,
                    "limit": variable.limit,
                    "cleaned_name": variable.cleaned_name,
                }
            )

        return (
            pd.DataFrame(variable_dict_list)
            .sort_values(by=["code"])
            .reset_index(drop=True)
        )

    @timer
    def stats(
        self,
        results: List[List[List[str]]],
        type_conversions: Dict[str, Any],
        geo_domains_queried: List[GeoDomain],
        column_headers: Dict[VariableCode, str],
        supported_geos: pd.DataFrame,
    ) -> pd.DataFrame:
        main_df = pd.DataFrame()

        mergeKeys = [domain.name for domain in geo_domains_queried]

        for result in results:
            df = pd.DataFrame(result[1:], columns=result[0])

            if main_df.empty:
                main_df = df
            else:
                df = df.drop(columns=["NAME"])  # type: ignore
                main_df = cast(pd.DataFrame, pd.merge(main_df, df, on=mergeKeys, how="inner"))  # type: ignore

        all_cols = main_df.columns.tolist()

        name_col, sorted_geo_cols, variable_cols = self._partition_stat_columns(
            column_headers,
            all_cols,
            supported_geos,
        )

        reorderedColumns = name_col + sorted_geo_cols + variable_cols

        return (
            main_df[reorderedColumns]  # type: ignore
            .astype(type_conversions)
            .rename(
                columns=column_headers if self._config.replace_column_headers else {}
            )
            .sort_values(by=sorted_geo_cols)
            .reset_index(drop=True)
        )

    def _partition_stat_columns(
        self,
        renamed_column_headers: Dict[VariableCode, str],
        df_columns: List[str],
        supported_geos: pd.DataFrame,
    ) -> Tuple[List[str], List[str], List[str]]:
        originalVariableHeaders = [str(col) for col in renamed_column_headers.keys()]
        nameHeader = ["NAME"]

        geo_cols = [
            col for col in df_columns if col not in nameHeader + originalVariableHeaders
        ]
        sorted_geo_cols = [
            domain.name
            for domain in self._sort_geo_domains(
                [GeoDomain(col) for col in geo_cols], supported_geos
            )
        ]

        variable_cols = [
            col for col in df_columns if col != "NAME" and col not in geo_cols
        ]

        return nameHeader, sorted_geo_cols, variable_cols

    def _sort_geo_domains(
        self,
        geo_domains: List[GeoDomain],
        supported_geos: pd.DataFrame,
    ) -> List[GeoDomain]:
        geoHierarchyMapping: List[Dict[str, str]] = (
            supported_geos[["name", "hierarchy"]].drop_duplicates().to_dict("records")
        )
        geoNameToHierarchy = {
            mapping["name"]: mapping["hierarchy"] for mapping in geoHierarchyMapping
        }

        return sorted(
            geo_domains, key=lambda geoDomain: geoNameToHierarchy[geoDomain.name]
        )



from abc import ABC, abstractmethod
from collections import OrderedDict
from typing import Any, Dict, Generator, List

from the_census._api.models import GeographyItem
from the_census._geographies.models import GeoDomain
from the_census._variables.models import Group, GroupVariable, VariableCode


class ICensusApiFetchService(ABC):
    """
    Interface for our API client, which will
    perform all fetches for the Census API
    """

    @abstractmethod
    def healthcheck(self) -> None:
        """
        makes sure that the API client is
        configured properly
        """
        ...

    @abstractmethod
    def geography_codes(
        self, for_domain: GeoDomain, in_domains: List[GeoDomain] = []
    ) -> List[List[str]]:
        """
        Gets all geography codes for a given location domain:

        Args:
            for_domain (GeoDomain): the domain you want to search. This must
            be a child of any provided `in_domains`, as specified in the API's
            geography hierarchy.

            in_domains (List[GeoDomain], optional): geography domains
            that may help specify the query (e.g., if you want to
            search all congressional districts in a particular state).
            Defaults to [].

        Returns:
            List[List[str]]: API response
        """
        ...

    @abstractmethod
    def group_data(self) -> Dict[str, Group]:
        """
        Retrieves data on available concept groups for a given dataset/survey

        Returns:
            Dict[str, Group]: Mapping of group ID to concept
        """
        ...

    @abstractmethod
    def supported_geographies(self) -> OrderedDict[str, GeographyItem]:
        """
        Retrieves all queryable geographies for a given dataset/survey

        Returns:
            OrderedDict[str, GeographyItem]: mapping between a geography
            and possible queries that can be made on it
        """
        ...

    @abstractmethod
    def variables_for_group(self, group: str) -> List[GroupVariable]:
        """
        Gets all queryable variables for a survey group concept.

        Args:
            group (str): The group's code

        Returns:
            List[GroupVariable]
        """
        ...

    @abstractmethod
    def all_variables(self) -> List[GroupVariable]:
        """
        Gets all variables. This may be costly

        Returns:
            List[GroupVariable]: all of the variables.
        """
        ...

    @abstractmethod
    def stats(
        self,
        variables_codes: List[VariableCode],
        for_domain: GeoDomain,
        in_domains: List[GeoDomain] = [],
    ) -> Generator[List[List[str]], None, None]:
        """
        Gets stats based on `variableCodes` for the geographies in question.
        Returns a generator, since we may need to make repeat API
        calls due to limits on the number of variables (50) that the API
        will accept to query at a time.

        Args:
            variables_codes (List[VariableCode])
            for_domain (GeoDomain)
            in_domains (List[GeoDomain], optional). Defaults to [].

        Yields:
            Generator[List[List[str]], None, None]
        """
        ...


class ICensusApiSerializationService(ABC):
    """
    Serialization layer between the raw API results & models
    """

    @abstractmethod
    def parse_group_variables(self, group_variables: Any) -> List[GroupVariable]:
        """
        Parses an API response for variable retrieval

        Args:
            group_variables (Any): JSON response

        Returns:
            List[GroupVariable]:
        """

        ...

    @abstractmethod
    def parse_supported_geographies(
        self, supported_geos_response: Any
    ) -> OrderedDict[str, GeographyItem]:
        """
        parse a supported geographies response from the census API:

        Args:
            supported_geos_response (Any)

        Returns:
            OrderedDict[str, GeographyItem]: mapping the geography title to its name and code
        """

        ...

    @abstractmethod
    def parse_groups(
        self, groups_res: Dict[str, List[Dict[str, str]]]
    ) -> Dict[str, Group]:
        """
        Parses a /groups.json response from the census API

        Args:
            groups_res (Dict[str, List[Dict[str, str]]])

        Returns:
            Dict[str, Group]
        """
        ...



class CensusDoesNotExistException(Exception):
    """
    Thrown during the healthcheck if the requested
    dataset/survey does not exist.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class InvalidQueryException(Exception):
    """
    Thrown if an invalid query was made to the Census API.
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class EmptyRepositoryException(Exception):
    """
    Thrown when trying to make a query on variables
    whose metadata is not in the repository
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class NoCensusApiKeyException(Exception):
    """
    Thrown when no Census API key is provided
    """
